"""
Tools for visualizing qcio objects in Jupyter Notebooks.

Design Decisions:
    - The view function is the top-level method for viewing all qcio objects. It can 
        accept one or many objects and will determine the appropriate viewer to use.
    - All functions that begin with "generate" will return a string of HTML. If users
        want to use this HTML to create a custom view, they can do so. If they want to
        display the HTML in a Jupyter Notebook, they can call display(HTML(html_string))
        after importing `from IPython.display import HTML, display`.
    - The basic layout for viewing results (all ProgramOutput objects) is a table of
        basic parameters followed by a structure viewer and results table or plot.
        DualProgramInputs add details for the subprogram.
        ----------------------------------------------------------------------------
        | Structure      | Success | Calculation Type | Program | Model | Keywords |
        ----------------------------------------------------------------------------
        |                                    |                                     |
        |      Structure Viewer (Optional)   |        Results Table or Plot        | 
        |                                    |                                     |
        ----------------------------------------------------------------------------
"""

import base64
import html
import io
import math
from contextlib import contextmanager
from itertools import zip_longest
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from qcio import (
    ConformerSearchResults,
    DistanceUnits,
    DualProgramInput,
    Files,
    OptimizationResults,
    ProgramInput,
    ProgramOutput,
    Results,
    SinglePointResults,
    Structure,
    constants,
)

# View dependencies
try:
    import matplotlib.pyplot as plt
    import py3Dmol as p3d
    from IPython.display import HTML, display
    from rdkit import Chem
    from rdkit.Chem import Draw
except ImportError as e:
    missing_packages = []
    if "matplotlib" in str(e):
        missing_packages.append("matplotlib")
    if "py3Dmol" in str(e):
        missing_packages.append("py3Dmol")
    if "IPython" in str(e):
        missing_packages.append("IPython")
    if "rdkit" in str(e):
        missing_packages.append("rdkit")
    raise ImportError(
        f"Missing dependencies: {', '.join(missing_packages)} required for the view "
        "module. Please install them using: pip install qcio[view]"
    )

DEFAULT_WIDTH: int = 600
"""The default width of the viewer in pixels."""
DEFAULT_HEIGHT: int = 450
"""The default height of the viewer in pixels."""


def generate_structure_viewer_html(
    *structs: Union["Structure", List["Structure"]],
    width: Optional[int] = None,
    height: Optional[int] = None,
    titles: Optional[List[str]] = None,
    subtitles: Optional[List[str]] = None,
    titles_extra: Optional[List[str]] = None,
    subtitles_extra: Optional[List[str]] = None,
    distances: Optional[List[Tuple[int, int]]] = None,
    distance_units: DistanceUnits = DistanceUnits.bohr,
    style: Optional[Dict] = None,
    show_indices: bool = False,
    same_viewer: bool = False,
    view_2d: bool = False,
    interval: int = 100,
) -> str:
    """
    Generate the core HTML viewer for a Structure or list of Structures using py3Dmol
    or 2D PNG images. These keywords may be passed to high level viewer functions.

    Args:
        structs: The Structure or list of Structures to visualize.
        width: The width of the viewer in pixels. Defaults to 600.
        height: The height of the viewer in pixels. Defaults to 450.
        titles: The titles to display above the viewer. Will default to the Structure
            name or SMILES if not provided.
        subtitles: The subtitles to display below the viewer.
        titles_extra: Extra text to display after the title. This is useful for adding
            additional context to a default title.
        subtitles_extra: Extra text to display after the subtitle. Useful for adding
            text after an autogenerated subtitle such as relative energy.
        distances: A list of tuples of atom indices to display distances between.
        distance_units: The units to display the distances in. Defaults to bohr. May
            pass 'angstrom' to display in angstroms.
        style: A dictionary of styles to apply to the viewer (for py3Dmol). Should be a
            dictionary for AtomStyleSpec <https://3dmol.org/doc/AtomStyleSpec.html>
        show_indices: Whether to show the atom indices in the viewer.
        view_2d: Whether to display the structure in 2D (PNG images).
        interval: The interval in milliseconds for the animation.

    Returns:
        str: The HTML string for the viewer.

    Raises:
        ValueError: If a list of Structures is passed and view_2d is True.
    """

    # Set defaults
    width = width or DEFAULT_WIDTH
    height = height or DEFAULT_HEIGHT

    titles = titles or []
    subtitles = subtitles or []
    titles_extra = titles_extra or []
    subtitles_extra = subtitles_extra or []

    # Start HTML with flex container
    html_parts = [
        '<div style="display: flex; flex-wrap: wrap; justify-content: center;">'
    ]

    if not view_2d:
        # Create the viewer
        if len(structs) == 1 or same_viewer:
            viewer = p3d.view(width=width, height=height, viewergrid=(1, 1))
        else:
            # Determine the number of rows needed for multiple structures
            rows = math.ceil(len(structs) / 2)
            viewer = p3d.view(
                width=width * 2,
                height=height * rows,
                linked=False,
                viewergrid=(rows, 2),
            )

    for i, (struct, title, subtitle, title_extra, subtitle_extra) in enumerate(
        zip_longest(structs, titles, subtitles, titles_extra, subtitles_extra)
    ):
        # Set the title and subtitle
        if isinstance(struct, list):
            name = struct[0].ids.name
            smiles = struct[0].ids.smiles
        else:
            name = struct.ids.name
            smiles = struct.ids.smiles

        title = f"{title or name or smiles or ''}"
        title_extra = f"{title_extra or ''}"
        subtitle = f"{subtitle or ''}"
        subtitle_extra = f"{subtitle_extra or ''}"

        if view_2d:
            if isinstance(struct, list):
                raise ValueError(
                    "Cannot display multiple 2D structures in one viewer. Do not submit"
                    " Structures in a list. If you want to view multiple structures, "
                    "unpack your list with *my_list_of_structures."
                )
            adjusted_width, adjusted_height = int(width * 0.75), int(height * 0.75)
            mol = Chem.MolFromSmiles(struct.ids.smiles or struct.to_smiles())
            img = Draw.MolToImage(mol, size=(adjusted_width, adjusted_height))
            buf = io.BytesIO()
            img.save(buf, format="PNG")
            img_str = base64.b64encode(buf.getvalue()).decode("utf-8")

            html_parts.append(
                f'<div style="margin: 10px; text-align: center; padding: 15px; width: '
                f'{adjusted_width}px; height: {adjusted_height + 60}px;">'
                f"<h2>{title} {title_extra}</h2>"
                f'<img src="data:image/png;base64,{img_str}" width="{adjusted_width}" '
                f'height="{adjusted_height}"/>'
                f"<h4>{subtitle} {subtitle_extra}</h4>"
                f"</div>"
            )
        else:
            if same_viewer:
                grid = (0, 0)
            else:
                grid = divmod(i, 2)

            if isinstance(struct, list):  # Animate lists of structures
                combined_xyz = "".join(s.to_xyz() for s in struct)
                viewer.addModelsAsFrames(combined_xyz, "xyz", viewer=grid)
                viewer.animate({"loop": "forward", "interval": interval}, viewer=grid)
                # viewer.animate({"loop": "forward"}, viewer=grid)
            else:
                viewer.addModel(struct.to_xyz(), "xyz", viewer=grid)

            viewer.addLabel(
                f"{title} {title_extra}",
                {
                    "position": {"x": width / 2, "y": 0, "z": 0},
                    "alignment": "topCenter",
                    "fontSize": 24,
                    "backgroundOpacity": 0,
                    "fontColor": "black",
                    "useScreen": True,
                },
                viewer=grid,
            )

            viewer.addLabel(
                f"{subtitle} {subtitle_extra}",
                {
                    "position": {"x": width / 2, "y": height, "z": 0},
                    "alignment": "bottomCenter",
                    "fontSize": 20,
                    "backgroundOpacity": 0,
                    "fontColor": "black",
                    "useScreen": True,
                },
                viewer=grid,
            )
            if distances:
                assert isinstance(
                    struct, Structure
                ), "Displaying distances for lists of structures is not yet implemented"
                for atom1, atom2 in distances:
                    a1_coords = struct.geometry_angstrom[atom1]
                    a2_coords = struct.geometry_angstrom[atom2]
                    # Add line between the two atoms
                    viewer.addLine(
                        {
                            "start": {
                                "x": a1_coords[0],
                                "y": a1_coords[1],
                                "z": a1_coords[2],
                            },
                            "end": {
                                "x": a2_coords[0],
                                "y": a2_coords[1],
                                "z": a2_coords[2],
                            },
                            "color": "red",
                            "linewidth": 2,
                        },
                        viewer=grid,
                    )
                    # Add a label showing the distance
                    midpoint = (a1_coords + a2_coords) / 2
                    distance = struct.distance(atom1, atom2, units=distance_units)
                    unit = "a₀"

                    if distance_units == DistanceUnits.angstrom:
                        unit = "Å"

                    viewer.addLabel(
                        f"{distance:.2f} {unit}",
                        {
                            "position": {
                                "x": midpoint[0],
                                "y": midpoint[1],
                                "z": midpoint[2],
                            },
                            "backgroundColor": "white",
                            "fontSize": 14,
                            "fontColor": "black",
                        },
                        viewer=grid,
                    )

            if show_indices:
                ang_geom = (
                    struct.geometry_angstrom
                    if not isinstance(struct, list)
                    else struct[0].geometry_angstrom
                )
                for j, coord in enumerate(ang_geom):
                    viewer.addLabel(
                        str(j),
                        {"position": {"x": coord[0], "y": coord[1], "z": coord[2]}},
                        viewer=grid,
                    )

    if not view_2d:
        viewer.setStyle(style or {"stick": {}, "sphere": {"scale": 0.3}})
        viewer.zoomTo()
        html_parts.append(f"{viewer.write_html()}")

    html_parts.append("</div>")
    return "".join(html_parts)


def generate_dictionary_string(dictionary: Dict[str, Any]) -> str:
    """
    Generate an HTML string displaying a dictionary without explicit row colors.

    Args:
        dictionary: The dictionary to display.

    Returns:
        str: A string of HTML displaying the dictionary with each key/value pair on a
            new line.
    """
    rows = "".join(
        f"<tr><td class='key'>{key}</td><td class='value'>{value}</td></tr>"
        for key, value in dictionary.items()
    )
    return f"<table class='inner-table'>{rows}</table>"


def generate_files_string(files: Dict[str, Union[str, bytes]]) -> str:
    """
    Generate an HTML string displaying a dictionary of files substituting <str> or
    <bytes> for the actual file data
    """
    # Generate the symbolic dictionary
    viewer_dict = {}
    for key, value in files.items():
        if isinstance(value, bytes):
            viewer_dict[key] = html.escape("<bytes>")
        elif isinstance(value, str):
            viewer_dict[key] = html.escape("<str>")

    return generate_dictionary_string(viewer_dict)


def generate_output_table(*prog_outputs: ProgramOutput) -> str:
    """
    Generate an HTML table displaying the basic parameters for ProgramOutput objects.

    Args:
        *prog_outputs: The ProgramOutput objects to display.

    Returns:
        str: A string of HTML displaying the ProgramOutput objects in a table.
    """
    styles = """
    <style>
        table {
            border-collapse: collapse;
            width: 100%;
        }
        th, td {
            border: 1px solid black;
            padding: 8px;
            text-align: left;
        }
        .inner-table {
            border: none;
            width: 100%;
        }
        .inner-table td {
            border: none;
            padding: 2px 5px;
        }
        .key, .value {
            text-align: left;
            white-space: nowrap;
        }
    </style>
    """

    table_header = """
    <table>
        <tr>
            <th>Structure</th>
            <th>Success</th>
            <th>Wall Time</th>
            <th>Calculation Type</th>
            <th>Program</th>
            <th>Model</th>
            <th>Keywords</th>
    """
    if any(po.input_data.files for po in prog_outputs):
        table_header += "<th>Input Files</th>"

    if any(isinstance(po.input_data, DualProgramInput) for po in prog_outputs):
        table_header += """
            <th>Subprogram</th>
            <th>Subprogram Model</th>
            <th>Subprogram Keywords</th>
        """
    table_header += "</tr>"

    table_rows = []
    for po in prog_outputs:
        success_style = (
            'style="color: green; font-weight: bold;"'
            if po.success
            else 'style="color: red; font-weight: bold;"'
        )
        base_row = f"""
        <tr>
            <td>{generate_dictionary_string(
                {
                    "charge": po.input_data.structure.charge, 
                    "multiplicity": po.input_data.structure.multiplicity, 
                    "name": po.input_data.structure.ids.name or "",
                })}</td>
            <td {success_style}>{po.success}</td>
            <td> {
                _format_time(po.provenance.wall_time) if po.provenance.wall_time 
                 else "No timing data"}</td>
            <td>{po.input_data.calctype.name}</td>
            <td>{f"{po.provenance.program} {po.provenance.program_version or ''}"}</td>
            <td>{generate_dictionary_string(
                po.input_data.model.model_dump(exclude=["extras"])
                ) if po.input_data.model else ""}</td>
            <td>{generate_dictionary_string(po.input_data.keywords)}</td>
        """
        if po.input_data.files:
            base_row += f"<td>{generate_files_string(po.input_data.files)}</td>"

        if isinstance(po.input_data, DualProgramInput):
            base_row += f"""
            <td>{po.input_data.subprogram}</td>
            <td>{po.input_data.subprogram_args.model}</td>
            <td>{generate_dictionary_string(po.input_data.subprogram_args.keywords)}</td>
            """
        base_row += "</tr>"
        table_rows.append(base_row)

    table_footer = "</table>"
    return styles + table_header + "\n".join(table_rows) + table_footer


def generate_optimization_plot(
    prog_output: ProgramOutput, figsize=(6.4, 4.8), grid=True
) -> str:
    """
    Generate a plot of the energy optimization by cycle for a single ProgramOutput.

    Args:
        prog_output: The ProgramOutput to generate the plot for.
        figsize: The size of the figure in inches.
        grid: Whether to display grid lines on the plot.

    Returns:
        str: A string of HTML displaying the plot as a png image encoded in base64.
    """
    energies = prog_output.results.energies * constants.HARTREE_TO_KCAL_PER_MOL
    baseline_energy = energies[0]
    relative_energies = energies - baseline_energy
    last_is_nan = np.isnan(relative_energies[-1])

    if last_is_nan:
        try:
            delta_E = relative_energies[-2]
        except IndexError:
            # If there is only one energy point, set delta_E to nan
            delta_E = np.nan
    else:
        delta_E = relative_energies[-1]

    fig, ax1 = plt.subplots(figsize=figsize)
    color = "tab:blue"
    ax1.set_xlabel("Optimization Cycle")
    ax1.set_ylabel("Relative Energy (kcal/mol)", color=color)
    ax1.plot(relative_energies, label="Energy", marker="o", color="green")
    if last_is_nan:
        ax1.plot(len(relative_energies) - 1, delta_E, marker="x", color="red")
    ax1.tick_params(axis="y", labelcolor=color)
    ax1.text(
        0.95,
        0.85,
        f"ΔE = {delta_E:.2f} kcal/mol",
        verticalalignment="top",
        horizontalalignment="right",
        transform=ax1.transAxes,
        color="black",
        fontsize=14,
    )
    plt.title("Energy Optimization by Cycle", pad=20)
    ax1.legend(loc="upper right")
    fig.tight_layout(rect=[0, 0, 1, 0.95])

    buf = io.BytesIO()
    plt.savefig(buf, format="png", bbox_inches="tight")
    buf.seek(0)
    image_base64 = base64.b64encode(buf.read()).decode("utf-8")
    buf.close()
    plt.close(fig)  # Close the figure to avoid duplicate plots
    return (
        f'<img src="data:image/png;base64,{image_base64}" alt="Energy Optimization by '
        f'Cycle" style="width: 100%; max-width: {DEFAULT_WIDTH}px;">'
    )


@contextmanager
def _numpy_print_options(*args, **kwargs):
    """
    Temporarily set numpy print options.

    Args:
        *args: Arguments for numpy print options.
        **kwargs: Keyword arguments for numpy print options.
    """
    original_options = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    try:
        yield
    finally:
        np.set_printoptions(**original_options)


def _not_empty(value) -> bool:
    """
    Check if a value is empty. Accommodates numpy arrays.

    Args:
        value: The value to check.

    Returns:
        bool: True if the value is not empty, False otherwise.
    """
    if isinstance(value, np.ndarray):
        return bool(value.any())
    return bool(value)


def _format_time(seconds_float: float) -> str:
    """Format a time in seconds to a human-readable string."""
    hours = int(seconds_float // 3600)
    minutes = int((seconds_float % 3600) // 60)
    seconds = seconds_float % 60
    if hours > 0:
        formatted_time = f"{hours:02}h:{minutes:02}m:{seconds:05.2f}s"
    elif minutes > 0:
        formatted_time = f"{minutes:02}m:{seconds:05.2f}s"
    else:
        formatted_time = f"{seconds:05.2f}s"

    return formatted_time


def generate_results_table(results: Files) -> str:
    """
    Generate an HTML table displaying the results.

    Args:
        results: The Results object to display.

    Returns:
        str: A string of HTML displaying the results in a table.
    """

    rows = ""
    for key, value in results.__dict__.items():
        if _not_empty(value):
            if key != "files":
                with _numpy_print_options(
                    threshold=10,
                    formatter={"float_kind": "{:0.2e}".format},
                ):
                    rows += f"<tr><td>{key}</td><td>{value}</td></tr>"

    # Add the files to the bottom table
    if _not_empty(results.files):
        rows += (
            f"<tr><td>Files</td><td>{generate_files_string(results.files)}</td></tr>"
        )

    return f"""
    <table style="width: 100%; border-collapse: collapse; text-align: left;">
        <thead>
            <tr>
                <th style="border: 1px solid black; padding: 8px;">Result</th>
                <th style="border: 1px solid black; padding: 8px;">Value</th>
            </tr>
        </thead>
        <tbody>
            {rows}
        </tbody>
    </table>
    """


def structures(
    *structs: Union[Structure, List[Structure]],
    **kwargs,
) -> None:
    """
    Display one or many Structure objects.

    Args:
        *structs: The Structure objects or list of objects to display. If a list is
            passed the structures will be animated.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the structures in the Jupyter Notebook.
    """
    display(HTML(generate_structure_viewer_html(*structs, **kwargs)))


def program_outputs(
    *prog_outputs: ProgramOutput[Union[ProgramInput, DualProgramInput], Results],
    animate: bool = True,
    struct_viewer: bool = True,
    **kwargs,
) -> None:
    """
    Display one or many ProgramOutput objects.

    Args:
        *prog_outputs: The ProgramOutput objects to display.
        animate: Whether to animate the structure if it is an optimization.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the ProgramOutput objects in the Jupyter Notebook.
    """

    width = kwargs.get("width", DEFAULT_WIDTH)
    height = kwargs.get("height", DEFAULT_HEIGHT)

    for i, po in enumerate(prog_outputs):
        final_html = []
        final_html.append(generate_output_table(po))

        if isinstance(po.results, ConformerSearchResults):
            structures = [po.input_data.structure] + po.results.conformers
            titles_extra = ["Initial Structure"] + [
                f"Conformer {i}" for i in range(len(po.results.conformers))
            ]
            subtitles = ["Rel Energy: Unknown"] + [
                f"Rel Energy: +{re * constants.HARTREE_TO_KCAL_PER_MOL:.3f} kcal/mol"
                for re in po.results.conformer_energies_relative
            ]
            conf_viewer = generate_structure_viewer_html(
                *structures, titles_extra=titles_extra, subtitles=subtitles, **kwargs
            )
            final_html.append(conf_viewer)
            display(HTML("".join(final_html)))

        else:
            # Create structure viewer
            if not struct_viewer:
                structure_html = "struct_viewer = False"

            else:
                titles_extra = kwargs.pop("titles_extra", [])
                try:
                    title_extra = titles_extra[i]
                except IndexError:
                    title_extra = ""

                # Determine the Structure to use
                if isinstance(po.results, OptimizationResults):
                    for_viewer: Union[Structure, List[Structure]]
                    if animate:
                        for_viewer = po.results.structures
                    else:
                        for_viewer = po.results.final_structure
                        title_extra += " (Final Structure)"

                elif isinstance(po.results, SinglePointResults):
                    for_viewer = po.input_data.structure

                elif isinstance(po.results, Files):
                    for_viewer = po.input_data.structure
                else:
                    raise NotImplementedError(
                        f"Viewing of {type(po.results)} is not yet implemented."
                    )

                structure_html = generate_structure_viewer_html(
                    for_viewer,
                    titles_extra=[title_extra],
                    **kwargs,
                )

            # Create results table or plot
            if isinstance(po.results, OptimizationResults):
                results_html = generate_optimization_plot(
                    po, figsize=(width / 100, height / 100)
                )
            else:
                results_html = generate_results_table(po.results)

            final_html.append(
                f"""
        <div style="text-align: center;">
            <div style="display: flex; align-items: center; justify-content: 
                space-around;">
                <div style="text-align: center; margin-right: 20px; flex: 1;">
                    <div style="display: inline-block; text-align: center;">
                        {structure_html}
                    </div>
                </div>
                <div style="width: {width}px; height: {height}px; text-align: center; 
                    margin-left: 20px; flex: 1; overflow: auto;">
                    {results_html}
                </div>
            </div>
        </div>
                """
            )

            display(HTML("".join(final_html)))


def view(
    *objs: Union[ProgramOutput, Structure, List[Structure]],
    **kwargs,
) -> None:
    """
    Top level method for viewing all qcio objects. This should be the only method you
    need to use to view any qcio object.

    Args:
        *objs: The ProgramOutput or Structure objects to view. May pass one or more
            objects or one or more lists of Structure objects.
        **kwargs: Additional keyword arguments to pass to the viewer functions.

    Returns:
        None. Displays the qcio objects in the Jupyter Notebook.
    """
    if all([isinstance(o, Structure) for o in objs]) or all(
        isinstance(o, Structure) for lst in objs for o in lst
    ):
        structures(*objs, **kwargs)  # type: ignore
        # Handle all structures in a single viewer
        return

    for obj in objs:
        if isinstance(obj, Structure) or isinstance(obj, list):
            structures(*objs, **kwargs)  # type: ignore

        elif isinstance(obj, ProgramOutput):
            program_outputs(obj, **kwargs)

        else:
            raise NotImplementedError(f"Viewing of {type(obj)} is not implemented.")



"""Physical constants and periodic table from NIST CODATA 2022 and PubChem. All values
must have a reference to their source.

References:
    - <https://physics.nist.gov/cuu/Constants/Table/allascii.txt>
    - PubChem periodic table JSON: <https://pubchem.ncbi.nlm.nih.gov/periodic-table/>
"""

import csv
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel

BOHR_TO_ANGSTROM = 0.529177210544
"""Conversion factor from Bohr to Angstrom. 

<https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0>
"""

ANGSTROM_TO_BOHR = 1 / BOHR_TO_ANGSTROM
"""Conversion factor from Angstrom to Bohr.

<https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0>
"""

HARTREE_TO_JOULE = 4.3597447222060e-18
"""Conversion factor from Hartree to Joule. 

<https://physics.nist.gov/cgi-bin/cuu/Value?na>
"""

AVOGADRO_NUMBER = 6.02214076e23
"""Avogadro's number. 

<https://physics.nist.gov/cuu/Constants/Table/allascii.txt>
"""

KCAL_TO_JOULE = 4.184e3
"""Conversion factor from kilocalorie to Joule.
Using kcal(th) as Therochemical calorie is the unit used in quantum chemistry.

<https://www.nist.gov/pml/special-publication-811/nist-guide-si-appendix-b-conversion-factors/nist-guide-si-appendix-b8>
"""

HARTREE_TO_KCAL_PER_MOL = HARTREE_TO_JOULE / KCAL_TO_JOULE * AVOGADRO_NUMBER
"""Conversion factor from Hartree to kcal/mol."""

BOLTZMANN_CONSTANT = 1.380649e-23
"""Boltzmann constant in J/K.
<https://physics.nist.gov/cuu/Constants/Table/allascii.txt>
"""

_DATA_DIR = Path(__file__).parent / "constants_data"


class Atom(BaseModel):
    """Atom data model."""

    symbol: str
    number: int
    name: str
    mass: float
    group: Optional[int]
    period: int
    block: str
    electron_config: str


class PeriodicTable:
    """Periodic table data model. Data source and URL should be set by class methods."""

    # Should be set by @classmethod
    data_source: str = ""
    data_url: str = ""

    @classmethod
    def from_pubchem(cls):
        """Create a periodic table from PubChem data."""
        instance = cls()
        instance.data_source = "PubChem"
        instance.data_url = "https://pubchem.ncbi.nlm.nih.gov/periodic-table/"

        group_and_period_data = {}

        with open(_DATA_DIR / "group_period.csv", mode="r") as file:
            reader = csv.DictReader(file)
            for row in reader:
                group_and_period_data[row["Symbol"]] = {
                    "group": int(row["Group"]) if row["Group"] else None,
                    "period": int(row["Period"]),
                }

        with open(_DATA_DIR / "pubchem.csv", mode="r") as file:
            reader = csv.DictReader(file)
            for row in reader:
                symbol = row["Symbol"]
                atom = Atom(
                    symbol=symbol,
                    number=int(row["AtomicNumber"]),
                    name=row["Name"],
                    mass=float(row["AtomicMass"]),
                    group=group_and_period_data[symbol]["group"],
                    period=group_and_period_data[symbol]["period"],
                    block=row["GroupBlock"],
                    electron_config=row["ElectronConfiguration"],
                )
                setattr(instance, symbol, atom)
        return instance

    def group(self, group_number: int) -> List[Atom]:
        """Return all atoms in a group."""
        assert 1 <= group_number <= 18, "Group number must be between 1 and 18."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.group == group_number
        ]

    def period(self, period_number: int) -> List[Atom]:
        """Return all atoms in a period."""
        assert 1 <= period_number <= 7, "Period number must be between 1 and 7."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.period == period_number
        ]

    def number(self, number: int) -> Atom:
        """Return an atom by atomic number."""
        for atom in self.__dict__.values():
            if isinstance(atom, Atom) and atom.number == number:
                return atom
        raise ValueError(f"No atom with atomic number {number}.")


# ruff: noqa
periodic_table = PeriodicTable.from_pubchem()
"""Periodic table data from PubChem.

    Attributes:
        data_source (str): Data source.
        data_url (str): Data URL.

    Example:
        ```python
        >>> from qcio.constnats import periodic_table as pt
        
        >>> pt.Ni
        Atom(symbol='Ni', number=28, name='Nickel', mass=58.6934, group=10, period=4, block='d', electron_config='[Ar] 3d8 4s2')

        >>> pt.group(1)
        [Atom(symbol='H', number=1, name='Hydrogen', mass=1.00784, group=1, period=1, block='s', electron_config='1s1'),
        Atom(symbol='Li', number=3, name='Lithium', mass=6.94, group=1, period=2, block='s', electron_config='[He] 2s1'),
        Atom(symbol='Na', number=11, name='Sodium', mass=22.98976928, group=1, period=3, block='s', electron_config='[Ne] 3s1'),
        Atom(symbol='K', number=19, name='Potassium', mass=39.0983, group=1, period=4, block='s', electron_config='[Ar] 4s1'),
        Atom(symbol='Rb', number=37, name='Rubidium', mass=85.4678, group=1, period=5, block='s', electron_config='[Kr] 5s1'),
        Atom(symbol='Cs', number=55, name='Cesium', mass=132.90545196, group=1, period=6, block='s', electron_config='[Xe] 6s1'),
        Atom(symbol='Fr', number=87, name='Francium', mass=223.0, group=1, period=7, block='s', electron_config='[Rn] 7s1')]
        ```
"""
# ruff: noqa: enable

# https://depts.washington.edu/eooptic/linkfiles/dielectric_chart%5B1%5D.pdf
_SOLVENTS_DIELECTRIC = {
    "acetic acid": 6.15,
    "acetone": 20.7,
    "acetonitrile": 37.5,
    "anisole": 4.33,
    "benzene": 2.27,
    "bromobenzene": 5.17,
    "carbon disulfide": 2.6,
    "carbon tetrachloride": 2.24,
    "chlorobenzene": 5.62,
    "chloroform": 4.81,
    "cyclohexane": 2.02,
    "dibutyl ether": 3.1,
    "o -dichlorobenzene": 9.93,
    "1,2-dichloroethane": 10.36,
    "dichloromethane": 8.93,
    "diethylamine": 3.6,
    "diethyl ether": 4.33,
    "1,2-dimethoxyethane": 7.2,
    "n,n -dimethylacetamide": 37.8,
    "n,n -dimethylformamide": 36.7,
    "dimethyl sulfoxide": 46.7,
    "1,4-dioxane": 2.25,
    "ethanol": 24.5,
    "ethyl acetate": 6.02,
    "ethyl benzoate": 6.02,
    "formamide": 111,
    "hexamethylphosphoramide": 30,
    "isopropyl alcohol": 17.9,
    "methanol": 32.7,
    "2-methyl-2-propanol": 10.9,
    "nitrobenzene": 34.82,
    "nitromethane": 35.87,
    "pyridine": 12.4,
    "tetrahydrofuran": 7.58,
    "toluene": 2.38,
    "trichloroethylene": 3.4,
    "triethylamine": 2.42,
    "trifluoroacetic acid": 8.55,
    "2,2,2-trifluoroethanol": 8.55,
    "water": 80.1,
    "o -xylene": 2.57,
}



"""Utility functions for the models module."""

import importlib
import warnings
from typing import TYPE_CHECKING, Any, Dict

import numpy as np

from ..constants import ANGSTROM_TO_BOHR
from ..constants import periodic_table as pt

if TYPE_CHECKING:
    from qcio.models.structure import Structure


def deprecated_function(new_name: str):
    """Notify users that a function is deprecated and will be removed in the future.
    Works equally well on functions and methods.

    Usage:
        @deprecated("new_name")
        def old_name():
            pass
    """

    def decorator(func):
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} is deprecated and will be removed in future "
                f"versions. Please use {new_name} instead.",
                FutureWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


def deprecated_class(new_name: str):
    """Notify users that a class is deprecated and will be removed in the future.
    Works only on Pydantic models since it relies on the model_post_init method.

    Usage:
        @deprecated_class("new_name")
        class OldName:
            pass
    """

    def class_decorator(cls):
        orig_init = cls.__init__

        def new_init(self, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and will be removed in a future "
                f"release. Please use '{new_name}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            orig_init(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    return class_decorator


def renamed_class(new_cls: Any):
    """Notify users that a class is deprecated and has been renamed.

    Usage:
        @renamed_class(NewClass)
        class OldClass:
            pass
    """

    def decorator(cls):
        def return_new_cls(cls, *args, **kwargs):
            warnings.warn(
                f"{cls.__name__} is deprecated and and will be removed in a future "
                f"release. Please use '{new_cls.__name__}' instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            return new_cls(*args, **kwargs)

        cls.__new__ = return_new_cls
        return cls

    return decorator


def _assert_module_installed(module: str):
    """Raise an error if the module is not installed."""
    try:
        importlib.import_module(module)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"The '{module}' module is required for this function. "
            f"Please install it with 'pip install qcio[{module}]' or "
            "'pip install qcio[all]'."
        )


def smiles_to_structure(
    smiles: str, program: str = "rdkit", force_field: str = "MMFF94s"
) -> Dict[str, Any]:
    """Convert a SMILES string to a Structure object in dictionary form.

    Args:
        smiles: The SMILES string to convert.
        method: The method to use for the conversion. Defaults to "MMFF94s".

    Returns:
        A dictionary representation of the Structure object.
    """
    # Remove newline characters if present
    smiles = smiles.strip()

    if program == "rdkit":
        if "." in smiles:
            raise ValueError(
                "Multiple molecules are not supported by RDKit. "
                "Please provide a single molecule or use openbabel for the program."
            )
        _assert_module_installed(program)
        from rdkit import Chem
        from rdkit.Chem import AllChem

        # Convert SMILES to RDKit Mol object
        mol = Chem.MolFromSmiles(smiles)
        assert mol is not None, f"Failed to convert SMILES to RDKit Mol: {smiles}"
        canonical_smiles = Chem.MolToSmiles(mol, canonical=True)
        mol = Chem.AddHs(mol)

        # Generate 3D coordinates
        AllChem.EmbedMolecule(mol, AllChem.ETKDG())
        # Optimize the molecule using the specified force field
        if force_field.upper() == "UFF":
            AllChem.UFFOptimizeMolecule(mol)
        elif force_field.upper() == "MMFF94":
            AllChem.MMFFOptimizeMolecule(mol, mmffVariant="MMFF94")  # type: ignore
        elif force_field.upper() == "MMFF94S":
            AllChem.MMFFOptimizeMolecule(mol, mmffVariant="MMFF94s")  # type: ignore
        else:
            raise ValueError(f"Unsupported force_field: {force_field}")

        # Get atom symbols
        atoms = [atom.GetSymbol() for atom in mol.GetAtoms()]  # type: ignore

        # Get atom positions
        conf = mol.GetConformer()
        geometry_angstrom = conf.GetPositions()
        geometry_bohr = geometry_angstrom * ANGSTROM_TO_BOHR

        # Get charge
        charge = Chem.GetFormalCharge(mol)

    elif program == "openbabel":
        _assert_module_installed(program)
        from openbabel import pybel

        # Convert SMILES to Open Babel Mol object
        mol = pybel.readstring("smi", smiles)
        assert mol is not None, f"Failed to convert SMILES to Open Babel Mol: {smiles}"

        # Add hydrogens
        mol.addh()  # type: ignore

        # Generate 3D coordinates and optimize
        mol.make3D(forcefield=force_field.lower(), steps=250)  # type: ignore

        # Get atom symbols
        atoms = [pt.number(atom.atomicnum).symbol for atom in mol.atoms]  # type: ignore

        # Get atom positions
        geometry_angstrom = np.array([atom.coords for atom in mol.atoms])  # type: ignore # noqa: E501
        geometry_bohr = geometry_angstrom * ANGSTROM_TO_BOHR

        # Get canonical SMILES
        canonical_smiles = mol.write("can").strip()  # type: ignore

        # Get charge
        charge = mol.charge  # type: ignore

    else:
        raise ValueError(f"Unsupported program: '{program}'.")

    return {
        "symbols": atoms,
        "geometry": geometry_bohr,
        "charge": charge,
        "identifiers": {
            "canonical_smiles": canonical_smiles,
            "smiles": smiles,
            "canonical_smiles_program": program,
        },
    }


def structure_to_smiles(
    structure: "Structure",
    *,
    program: str = "rdkit",
    hydrogens: bool = False,
) -> str:
    """Convert a Structure to a SMILES string.

    Args:
        structure: The Structure object to convert.
        program: The program to use for the conversion. Can be 'rdkit' or 'openbabel'.
        hydrogens: Whether to include hydrogens in the SMILES string.

    Returns:
        The SMILES string.
    """

    if program == "rdkit":
        _assert_module_installed(program)
        from rdkit import Chem
        from rdkit.Chem import rdDetermineBonds

        # Details: https://greglandrum.github.io/rdkit-blog/posts/2022-12-18-introducing-rdDetermineBonds.html  # noqa: E501
        # Create RDKit molecule
        mol = Chem.MolFromXYZBlock(structure.to_xyz())  # type: ignore

        # Use rdDetermineBonds to infer bond information
        rdDetermineBonds.DetermineBonds(mol, charge=structure.charge)

        # Remove hydrogens if necessary
        if not hydrogens:
            mol = Chem.RemoveHs(mol)

        return Chem.MolToSmiles(mol, canonical=True)

    elif program == "openbabel":
        _assert_module_installed(program)
        from openbabel import openbabel as ob
        from openbabel import pybel

        # Must remove data in second line for Open Babel
        xyz_lines = structure.to_xyz().splitlines()
        xyz_lines[1] = ""

        # Create Open Babel OBMol object
        mol = pybel.readstring("xyz", "\n".join(xyz_lines))

        # Assign charges
        partial_charges = mol.calccharges()

        # Check if the sum of the partial charges matches the total charge
        if sum(partial_charges) != structure.charge:
            raise ValueError(
                f"Charge mismatch. Open Babel: {sum(partial_charges)} vs Structure: "
                f"{structure.charge}"
            )

        # Set the formal charges on the atoms
        if sum(partial_charges) != 0:
            for atom, charge in zip(mol.atoms, partial_charges):
                atom.OBAtom.SetFormalCharge(int(round(charge)))

        # Ensure the total charge matches the structure
        if mol.charge != structure.charge:
            raise ValueError(
                f"Charge mismatch. Open Babel: {sum(partial_charges)} vs Structure: "
                f"{structure.charge}"
            )

        # Create an OBConversion object to handle output format
        conv = ob.OBConversion()
        conv.SetOutFormat("can")

        if hydrogens:
            conv.AddOption("h", ob.OBConversion.OUTOPTIONS)

        # Generate canonical SMILES with explicit hydrogens
        return conv.WriteString(mol.OBMol).strip()

    else:
        raise ValueError(f"Unsupported program: '{program}'.")



"""Utility functions for working with qcio objects."""

import json
from collections import Counter
from typing import List, Tuple, Union

import numpy as np
from pydantic import BaseModel

from .constants import ANGSTROM_TO_BOHR
from .models import Structure
from .models.utils import _assert_module_installed

# Helper Structures
water = Structure(
    symbols=["O", "H", "H"],
    geometry=np.array(
        [
            [0.0253397, 0.01939466, -0.00696322],
            [0.22889176, 1.84438441, 0.16251426],
            [1.41760224, -0.62610794, -1.02954938],
        ]
    ),
    charge=0,
    multiplicity=1,
    connectivity=[(0, 1, 1.0), (0, 2, 1.0)],
    identifiers={"name": "water"},
)


def json_dumps(obj: Union[BaseModel, List[BaseModel]]) -> str:
    """Serialization helper for lists of pydantic objects."""
    if isinstance(obj, list):
        return json.dumps([o.model_dump() for o in obj])
    return obj.model_dump_json()


def _rdkit_mol_from_structure(
    struct: Structure,
) -> "rdkit.Chem.Mol":  # type: ignore # noqa: F821
    """Create an RDKit molecule from a Structure object."""
    _assert_module_installed("rdkit")
    from rdkit import Chem

    # Create RDKit molecule
    mol = Chem.MolFromXYZBlock(struct.to_xyz())  # type: ignore

    if mol is None:
        raise ValueError("Failed create rdkit Molecule from xyz string.")

    # Ensure molecule has conformers
    if mol.GetNumConformers() == 0:
        raise ValueError("Molecule lacks 3D coordinates.")

    return mol


def rmsd(
    struct1: Structure, struct2: Structure, align: bool = True, numthreads: int = 1
) -> float:
    """
    Calculate the root mean square deviation between two structures in Angstrom.

    Args:
        struct1: The first structure.
        struct2: The second structure.
        align: Whether to align the structures before calculating the RMSD including
            atom renumbering. If True, rdkit will compute the optimal alignment of the
            two structures before calculating the RMSD. If False, the RMSD will be
            calculated without alignment.
        numthreads: The number of threads to use for the RMSD calculation. Applies only
            to the alignment step if `align=True`.


    Returns:
        The RMSD between the two structures in Angstroms.
    """
    _assert_module_installed("rdkit")
    from rdkit.Chem import rdDetermineBonds, rdMolAlign

    # Create RDKit molecules
    mol1 = _rdkit_mol_from_structure(struct1)
    mol2 = _rdkit_mol_from_structure(struct2)

    # Determine connectivity
    rdDetermineBonds.DetermineConnectivity(mol1, charge=struct1.charge)
    rdDetermineBonds.DetermineConnectivity(mol2, charge=struct2.charge)

    # Compute RMSD
    if align:
        rmsd = rdMolAlign.GetBestRMS(mol2, mol1, numThreads=numthreads)
    else:
        rmsd = rdMolAlign.CalcRMS(mol2, mol1)

    return rmsd


def align(
    struct: Structure, refstruct: Structure, reorder_atoms: bool = True
) -> Tuple[Structure, float]:
    """Return a new structure that is optimally aligned to the reference structure.

    Args:
        struct: The structure to align.
        refstruct: The reference structure.
        reorder_atoms: Reorder the atoms to match the reference structure. If False,
            the atoms will be aligned without changing their order.

    Returns:
        Tuple of the aligned structure and the RMSD in Angstroms.
    """
    _assert_module_installed("rdkit")
    from rdkit.Chem import rdDetermineBonds, rdMolAlign

    # Create RDKit molecules
    mol = _rdkit_mol_from_structure(struct)
    refmol = _rdkit_mol_from_structure(refstruct)

    # Determine connectivity
    rdDetermineBonds.DetermineConnectivity(mol, charge=struct.charge)
    rdDetermineBonds.DetermineConnectivity(refmol, charge=refstruct.charge)

    # Compute RMSD and align mol to refmol
    if reorder_atoms:
        rmsd, trnsfm_matrix, atm_map = rdMolAlign.GetBestAlignmentTransform(mol, refmol)  # type: ignore # noqa: E501
    else:
        rmsd, trnsfm_matrix = rdMolAlign.GetAlignmentTransform(mol, refmol)

    # Convert to homogeneous coordinates in Angstroms
    coords_homogeneous = np.hstack(
        [struct.geometry_angstrom, np.ones((struct.geometry.shape[0], 1))]
    )

    # Apply the transformation matrix
    transformed_coords_homogeneous = coords_homogeneous @ trnsfm_matrix.T

    # Extract the transformed 3D coordinates and convert to Bohr
    transformed_coords = transformed_coords_homogeneous[:, :3] * ANGSTROM_TO_BOHR

    # Reorder the atoms to match the reference structure
    if reorder_atoms:
        if Counter(struct.symbols) != Counter(refstruct.symbols):
            raise ValueError(
                "Structures must have the same number and type of atoms for "
                "`reorder_atoms=True` at this time. Pass "
                "`reorder_atoms=False` to align structures with different atom "
                "counts."
            )
        symbols = refstruct.symbols
        geometry = np.zeros((len(atm_map), 3))

        for probe_idx, ref_idx in atm_map:
            geometry[ref_idx] = transformed_coords[probe_idx]  # works

    # Otherwise, keep the original atom order
    else:
        symbols = struct.symbols
        geometry = transformed_coords

    return (
        Structure(
            symbols=symbols,
            geometry=geometry,
            charge=struct.charge,
            multiplicity=struct.multiplicity,
            connectivity=struct.connectivity,
            identifiers=struct.identifiers,
        ),
        rmsd,
    )
